@@ -31,13 +31,15 @@
 #include "libglnx/libglnx.h"
 
 #include "flatpak-builtins.h"
-#include "flatpak-utils.h"
-#include "flatpak-run.h"
+#include "flatpak-context-private.h"
+#include "flatpak-utils-private.h"
+#include "flatpak-run-private.h"
 
 static char *opt_command;
 static char *opt_require_version;
 static char **opt_extra_data;
 static char **opt_extensions;
+static char **opt_remove_extensions;
 static char **opt_metadata;
 static gboolean opt_no_exports;
 static int opt_extension_prio = G_MININT;
@@ -50,6 +52,7 @@ static GOptionEntry options[] = {
   { "no-exports", 0, 0, G_OPTION_ARG_NONE, &opt_no_exports, N_("Don't process exports"), NULL },
   { "extra-data", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_extra_data, N_("Extra data info") },
   { "extension", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_extensions, N_("Add extension point info"),  N_("NAME=VARIABLE[=VALUE]") },
+  { "remove-extension", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_remove_extensions, N_("Remove extension point info"),  N_("NAME") },
   { "extension-priority", 0, 0, G_OPTION_ARG_INT, &opt_extension_prio, N_("Set extension priority (only for extensions)"), N_("0") },
   { "sdk", 0, 0, G_OPTION_ARG_STRING, &opt_sdk, N_("Change the sdk used for the app"),  N_("SDK") },
   { "runtime", 0, 0, G_OPTION_ARG_STRING, &opt_runtime, N_("Change the runtime used for the app"),  N_("RUNTIME") },
@@ -63,7 +66,9 @@ export_dir (int           source_parent_fd,
             const char   *source_relpath,
             int           destination_parent_fd,
             const char   *destination_name,
-            const char   *required_prefix,
+            char        **allowed_prefixes,
+            char        **allowed_extensions,
+            gboolean      require_exact_match,
             GCancellable *cancellable,
             GError      **error)
 {
@@ -126,17 +131,34 @@ export_dir (int           source_parent_fd,
           g_autofree gchar *child_relpath = g_build_filename (source_relpath, dent->d_name, NULL);
 
           if (!export_dir (source_iter.fd, dent->d_name, child_relpath, destination_dfd, dent->d_name,
-                           required_prefix, cancellable, error))
+                           allowed_prefixes, allowed_extensions, require_exact_match,
+                           cancellable, error))
             return FALSE;
         }
       else if (S_ISREG (stbuf.st_mode))
         {
+          g_autofree gchar *name_without_extension = NULL;
+          int i;
+
           source_printable = g_build_filename (source_relpath, dent->d_name, NULL);
 
-
-          if (!flatpak_has_name_prefix (dent->d_name, required_prefix))
+          for (i = 0; allowed_extensions[i] != NULL; i++)
             {
-              g_print (_("Not exporting %s, wrong prefix\n"), source_printable);
+              if (g_str_has_suffix (dent->d_name, allowed_extensions[i]))
+                break;
+            }
+
+          if (allowed_extensions[i] == NULL)
+            {
+              g_print (_("Not exporting %s, wrong extension\n"), source_printable);
+              continue;
+            }
+
+          name_without_extension = g_strndup (dent->d_name, strlen (dent->d_name) - strlen (allowed_extensions[i]));
+
+          if (!flatpak_name_matches_one_wildcard_prefix (name_without_extension, (const char * const *) allowed_prefixes, require_exact_match))
+            {
+              g_print (_("Not exporting %s, non-allowed export filename\n"), source_printable);
               continue;
             }
 
@@ -177,7 +199,9 @@ static gboolean
 copy_exports (GFile        *source,
               GFile        *destination,
               const char   *source_prefix,
-              const char   *required_prefix,
+              char        **allowed_prefixes,
+              char        **allowed_extensions,
+              gboolean      require_exact_match,
               GCancellable *cancellable,
               GError      **error)
 {
@@ -187,17 +211,23 @@ copy_exports (GFile        *source,
   /* The fds are closed by this call */
   if (!export_dir (AT_FDCWD, flatpak_file_get_path_cached (source), source_prefix,
                    AT_FDCWD, flatpak_file_get_path_cached (destination),
-                   required_prefix, cancellable, error))
+                   allowed_prefixes, allowed_extensions, require_exact_match,
+                   cancellable, error))
     return FALSE;
 
   return TRUE;
 }
 
 static gboolean
-collect_exports (GFile *base, const char *app_id, GCancellable *cancellable, GError **error)
+collect_exports (GFile          *base,
+                 const char     *app_id,
+                 FlatpakContext *arg_context,
+                 GCancellable   *cancellable,
+                 GError        **error)
 {
   g_autoptr(GFile) files = NULL;
   g_autoptr(GFile) export = NULL;
+  int i;
   const char *paths[] = {
     "share/applications",                 /* Copy desktop files */
     "share/mime/packages",                /* Copy MIME Type files */
@@ -206,7 +236,7 @@ collect_exports (GFile *base, const char *app_id, GCancellable *cancellable, GEr
     "share/gnome-shell/search-providers", /* Search providers */
     NULL,
   };
-  int i;
+
 
   files = g_file_get_child (base, "files");
 
@@ -220,20 +250,35 @@ collect_exports (GFile *base, const char *app_id, GCancellable *cancellable, GEr
 
   for (i = 0; paths[i]; i++)
     {
-      g_autoptr(GFile) src = NULL;
-      src = g_file_resolve_relative_path (files, paths[i]);
+      const char * path = paths[i];
+      g_autoptr(GFile) src = g_file_resolve_relative_path (files, path);
+      g_auto(GStrv) allowed_prefixes = NULL;
+      g_auto(GStrv) allowed_extensions = NULL;
+      gboolean require_exact_match = FALSE;
+
+      if (!flatpak_get_allowed_exports (path, app_id, arg_context,
+                                        &allowed_extensions, &allowed_prefixes, &require_exact_match))
+        return flatpak_fail (error, "Unexpectedly not allowed to export %s", path);
+
       if (g_file_query_exists (src, cancellable))
         {
-          g_debug ("Exporting from %s", paths[i]);
+          g_debug ("Exporting from %s", path);
           g_autoptr(GFile) dest = NULL;
           g_autoptr(GFile) dest_parent = NULL;
-          dest = g_file_resolve_relative_path (export, paths[i]);
+          dest = g_file_resolve_relative_path (export, path);
           dest_parent = g_file_get_parent (dest);
-          g_debug ("Ensuring export/%s parent exists", paths[i]);
+          g_debug ("Ensuring export/%s parent exists", path);
           if (!flatpak_mkdir_p (dest_parent, cancellable, error))
             return FALSE;
-          g_debug ("Copying from files/%s", paths[i]);
-          if (!copy_exports (src, dest, paths[i], app_id, cancellable, error))
+          g_debug ("Copying from files/%s", path);
+          if (!copy_exports (src,
+                             dest,
+                             path,
+                             allowed_prefixes,
+                             allowed_extensions,
+                             require_exact_match,
+                             cancellable,
+                             error))
             return FALSE;
         }
     }
@@ -469,6 +514,13 @@ update_metadata (GFile *base, FlatpakContext *arg_context, gboolean is_runtime,
       g_key_file_set_string (keyfile, elements[0], elements[1], elements[2] ? elements[2] : "true");
     }
 
+  for (i = 0; opt_remove_extensions != NULL && opt_remove_extensions[i] != NULL; i++)
+    {
+      g_autofree char *groupname = g_strconcat (FLATPAK_METADATA_GROUP_PREFIX_EXTENSION, opt_remove_extensions[i], NULL);
+
+      g_key_file_remove_group (keyfile, groupname, NULL);
+    }
+
   for (i = 0; opt_extensions != NULL && opt_extensions[i] != NULL; i++)
     {
       g_auto(GStrv) elements = NULL;
@@ -568,7 +620,7 @@ flatpak_builtin_build_finish (int argc, char **argv, GCancellable *cancellable,
   if (!is_runtime)
     {
       g_debug ("Collecting exports");
-      if (!collect_exports (base, id, cancellable, error))
+      if (!collect_exports (base, id, arg_context, cancellable, error))
         return FALSE;
     }
 
@@ -602,7 +654,7 @@ flatpak_complete_build_finish (FlatpakCompletion *completion)
     case 1: /* DIR */
       flatpak_complete_options (completion, global_entries);
       flatpak_complete_options (completion, options);
-      flatpak_context_complete (arg_context, completion);
+      flatpak_complete_context (completion);
 
       flatpak_complete_dir (completion);
       break;
