@@ -29,11 +29,11 @@
 #include <glib/gi18n.h>
 
 #include "libglnx/libglnx.h"
-#include "flatpak-document-dbus.h"
+#include "flatpak-document-dbus-generated.h"
 
 #include "flatpak-builtins.h"
-#include "flatpak-utils.h"
-#include "flatpak-run.h"
+#include "flatpak-utils-private.h"
+#include "flatpak-run-private.h"
 
 static GOptionEntry options[] = {
   { NULL }
