@@ -32,7 +32,7 @@
 #include <unistd.h>
 #include <gio/gunixfdlist.h>
 
-#include <glib/gi18n.h>
+#include <glib/gi18n-lib.h>
 
 #include <gio/gio.h>
 #include "libglnx/libglnx.h"
