@@ -21,7 +21,7 @@
 #include "config.h"
 
 #include <stdio.h>
-#include <glib/gi18n.h>
+#include <glib/gi18n-lib.h>
 #include <libsoup/soup.h>
 
 #include "flatpak-transaction-private.h"
@@ -44,6 +44,14 @@
  * This means you should either handle the signals directly (say, by doing blocking console
  * interaction, or by just returning without interaction), or run the operation in a separate
  * thread and do your own forwarding to the GUI thread.
+ *
+ * Despite the name, a FlatpakTransaction is more like a batch operation than a transaction
+ * in the database sense. Individual operations are carried out sequentially, and are atomic.
+ * They become visible to the system as they are completed. When an error occurs, already
+ * completed operations are not rolled back.
+ *
+ * The FlatpakTransaction API is threadsafe in the sense that it is safe to run two
+ * transactions at the same time, in different threads (or processes).
  */
 
 /* This is an internal-only element of FlatpakTransactionOperationType */
@@ -171,8 +179,25 @@ bundle_data_free (BundleData *data)
 
 static guint progress_signals[LAST_SIGNAL] = { 0 };
 
+/**
+ * SECTION:flatpak-transaction-progress
+ * @Title: FlatpakTransactionProgress
+ * @Short_description: Progress of an operation
+ *
+ * FlatpakTransactionProgress is an object that represents the progress
+ * of a single operation in a transaction. You obtain a FlatpakTransactionProgress
+ * with the #FlatpakTransaction::new-operation signal.
+ */
+
 G_DEFINE_TYPE (FlatpakTransactionProgress, flatpak_transaction_progress, G_TYPE_OBJECT)
 
+/**
+ * flatpak_transaction_progress_set_update_frequency:
+ * @self: a #FlatpakTransactionProgress
+ * @update_frequency: the update frequency, in milliseconds
+ *
+ * Sets how often progress should be updated.
+ */
 void
 flatpak_transaction_progress_set_update_frequency (FlatpakTransactionProgress *self,
                                                    guint                       update_frequency)
@@ -180,19 +205,42 @@ flatpak_transaction_progress_set_update_frequency (FlatpakTransactionProgress *s
   g_object_set_data (G_OBJECT (self->ostree_progress), "update-frequency", GUINT_TO_POINTER (update_frequency));
 }
 
-
+/**
+ * flatpak_transaction_progress_get_status:
+ * @self: a #FlatpakTransactionProgress
+ *
+ * Gets the current status string
+ * 
+ * Returns: (transfer none): the current status 
+ */
 char *
 flatpak_transaction_progress_get_status (FlatpakTransactionProgress *self)
 {
   return g_strdup (self->status);
 }
 
+/**
+ * flatpak_transaction_progress_get_is_estimating:
+ * @self: a #FlatpakTransactionProgress
+ *
+ * Gets whether the progress is currently estimating
+ *
+ * Returns: whether we're estimating
+ */
 gboolean
 flatpak_transaction_progress_get_is_estimating (FlatpakTransactionProgress *self)
 {
   return self->estimating;
 }
 
+/**
+ * flatpak_transaction_progress_get_progress:
+ * @self: a #FlatpakTransactionProgress
+ *
+ * Gets the current progress.
+ * 
+ * Returns: the current progress, as an integer between 0 and 100
+ */
 int
 flatpak_transaction_progress_get_progress (FlatpakTransactionProgress *self)
 {
@@ -398,6 +446,16 @@ dir_ref_is_installed (FlatpakDir *dir, const char *ref, char **remote_out, GVari
   return TRUE;
 }
 
+/**
+ * SECTION:flatpak-transaction-operation
+ * @Title: FlatpakTransactionOperation
+ * @Short_description: Operation in a transaction
+ *
+ * FlatpakTransactionOperation is an object that represents a single operation
+ * in a transaction. You receive a FlatpakTransactionOperation object with the
+ * #FlatpakTransaction::new-operation signal.
+ */
+
 G_DEFINE_TYPE (FlatpakTransactionOperation, flatpak_transaction_operation, G_TYPE_OBJECT)
 
 static void
@@ -462,24 +520,56 @@ flatpak_transaction_operation_new (const char                     *remote,
   return self;
 }
 
+/**
+ * flatpak_transaction_operation_get_operation_type:
+ * @self: a #FlatpakTransactionOperation
+ *
+ * Gets the type of the operation.
+ *
+ * Returns: the type of operation, as #FlatpakTransactionOperationType
+ */
 FlatpakTransactionOperationType
 flatpak_transaction_operation_get_operation_type (FlatpakTransactionOperation *self)
 {
   return self->kind;
 }
 
+/**
+ * flatpak_transaction_operation_get_ref:
+ * @self: a #FlatpakTransactionOperation
+ *
+ * Gets the ref that the operation applies to.
+ *
+ * Returns: (transfer none): the ref
+ */ 
 const char *
 flatpak_transaction_operation_get_ref (FlatpakTransactionOperation *self)
 {
   return self->ref;
 }
 
+/**
+ * flatpak_transaction_operation_get_remote:
+ * @self: a #FlatpakTransactionOperation
+ *
+ * Gets the remote that the operation applies to.
+ *
+ * Returns: (transfer none): the remote
+ */ 
 const char *
 flatpak_transaction_operation_get_remote (FlatpakTransactionOperation *self)
 {
   return self->remote;
 }
 
+/**
+ * flatpak_transaction_operation_type_to_string:
+ * @kind: a #FlatpakTransactionOperationType
+ *
+ * Converts the operation type to a string.
+ *
+ * Returns: (transfer none): a string representing @kind
+ */
 const char *
 flatpak_transaction_operation_type_to_string (FlatpakTransactionOperationType kind)
 {
@@ -508,24 +598,71 @@ flatpak_transaction_operation_get_bundle_path (FlatpakTransactionOperation *self
   return self->bundle;
 }
 
+/**
+ * flatpak_transaction_operation_get_commit:
+ * @self: a #FlatpakTransactionOperation
+ *
+ * Gets the commit ID for the operation.
+ *
+ * This information is available when the transaction is resolved,
+ * i.e. when #FlatpakTransaction::ready is emitted.
+ *
+ * Returns: (transfer none): the commit ID
+ */
 const char *
 flatpak_transaction_operation_get_commit (FlatpakTransactionOperation *self)
 {
   return self->resolved_commit;
 }
 
+/**
+ * flatpak_transaction_operation_get_metadata:
+ * @self: a #FlatpakTransactionOperation
+ *
+ * Gets the metadata that will be applicable when the
+ * operation is done.
+ *
+ * This can be compared to the current metadata returned
+ * by flatpak_transaction_operation_get_old_metadata()
+ * to find new required permissions and similar changes.
+ *
+ * This information is available when the transaction is resolved,
+ * i.e. when #FlatpakTransaction::ready is emitted.
+ *
+ * Returns: (transfer none): the metadata #GKeyFile
+ */
 GKeyFile *
 flatpak_transaction_operation_get_metadata (FlatpakTransactionOperation *self)
 {
   return self->resolved_metakey;
 }
 
+/**
+ * flatpak_transaction_operation_get_old_metadata:
+ * @self: a #FlatpakTransactionOperation
+ *
+ * Gets the metadata current metadata for the ref that @self works on.
+ * Also see flatpak_transaction_operation_get_metadata().
+ *
+ * This information is available when the transaction is resolved,
+ * i.e. when #FlatpakTransaction::ready is emitted.
+ *
+ * Returns: (transfer none): the old metadata #GKeyFile
+ */
 GKeyFile *
 flatpak_transaction_operation_get_old_metadata (FlatpakTransactionOperation *self)
 {
   return self->resolved_old_metakey;
 }
 
+/**
+ * flatpak_transaction_is_empty:
+ * @self: a #FlatpakTransaction
+ *
+ * Returns whether the transaction contains any operations.
+ *
+ * Returns: %TRUE if the transaction is empty
+ */
 gboolean
 flatpak_transaction_is_empty (FlatpakTransaction *self)
 {
@@ -643,6 +780,11 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
   object_class->get_property = flatpak_transaction_get_property;
   object_class->set_property = flatpak_transaction_set_property;
 
+  /**
+   * FlatpakTransaction:installation:
+   *
+   * The installation that the transaction operates on.
+   */
   g_object_class_install_property (object_class,
                                    PROP_INSTALLATION,
                                    g_param_spec_object ("installation",
@@ -656,6 +798,9 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
    * @object: A #FlatpakTransaction
    * @operation: The new #FlatpakTransactionOperation
    * @progress: A #FlatpakTransactionProgress for @operation
+   *
+   * The ::new-operation signal gets emitted during the execution of
+   * the transaction when a new operation is beginning.
    */
   signals[NEW_OPERATION] =
     g_signal_new ("new-operation",
@@ -673,6 +818,9 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
    * @error: A #GError
    * @details: A #FlatpakTransactionErrorDetails with details about the error
    *
+   * The ::operation-error signal gets emitted when an error occurs during the
+   * execution of the transaction.
+   *
    * Returns: the %TRUE to contine transaction, %FALSE to stop
    */
   signals[OPERATION_ERROR] =
@@ -690,6 +838,8 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
    * @operation: The #FlatpakTransactionOperation which finished
    * @result: A #FlatpakTransactionResult giving details about the result
    *
+   * The ::operation-done signal gets emitted during the execution of
+   * the transaction when an operation is finished.
    */
   signals[OPERATION_DONE] =
     g_signal_new ("operation-done",
@@ -707,6 +857,9 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
    * @runtime_ref: The ref we are looking for
    * @remotes: the remotes that has the ref, sorted in prio order
    *
+   * The ::choose-remote-for-ref signal gets emitted when a
+   * remote needs to be selected during the execution of the transaction.
+   *
    * Returns: the index of the remote to use, or -1 to not pick one (and fail)
    */
   signals[CHOOSE_REMOTE_FOR_REF] =
@@ -717,12 +870,16 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
                   NULL, NULL,
                   NULL,
                   G_TYPE_INT, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRV);
+
   /**
    * FlatpakTransaction::end-of-lifed:
    * @object: A #FlatpakTransaction
    * @ref: The ref we are installing
    * @reason: The eol reason, or %NULL
    * @rebase: The new name, if rebased, or %NULL
+   *
+   * The ::end-of-lifed signal gets emitted when a ref is found to
+   * be marked as end-of-life during the execution of the transaction.
    */
   signals[END_OF_LIFED] =
     g_signal_new ("end-of-lifed",
@@ -732,14 +889,17 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
                   NULL, NULL,
                   NULL,
                   G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
+
   /**
    * FlatpakTransaction::ready:
    * @object: A #FlatpakTransaction
    *
-   * This is is emitted when all the refs involved in the operation have been
-   * resolved to commits. At this point flatpak_transaction_get_operations()
+   * The ::ready signal is emitted when all the refs involved in the operation
+   * have been resolved to commits. At this point flatpak_transaction_get_operations()
    * will return all the operations that will be executed as part of the
-   * transaction. If this returns FALSE, the operation is aborted.
+   * transaction.
+   *
+   * Returns: %TRUE to carry on with the transaction, %FALSE to abort
    */
   signals[READY] =
     g_signal_new ("ready",
@@ -749,6 +909,7 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
                   signal_accumulator_false_abort, NULL,
                   NULL,
                   G_TYPE_BOOLEAN, 0);
+
   /**
    * FlatpakTransaction::add-new-remote:
    * @object: A #FlatpakTransaction
@@ -757,9 +918,11 @@ flatpak_transaction_class_init (FlatpakTransactionClass *klass)
    * @suggested_remote_name: The suggested remote name
    * @url: The repo url
    *
-   * As part of the transaction, it is required or recommended
-   * that a new remote is added, for the reason described in @reason.
-   * Return %TRUE to add it.
+   * The ::add-new-remote signal gets emitted if, as part of the transaction,
+   * it is required or recommended that a new remote is added, for the reason
+   * described in @reason.
+   *
+   * Returns: %TRUE to add the remote
    */
   signals[ADD_NEW_REMOTE] =
     g_signal_new ("add-new-remote",
@@ -836,6 +999,14 @@ flatpak_transaction_new_for_installation (FlatpakInstallation *installation,
                          NULL);
 }
 
+/**
+ * flatpak_transaction_set_no_pull:
+ * @self: a #FlatpakTransaction
+ * @no_pull: whether to avoid pulls
+ *
+ * Sets whether the transaction should operate only on locally
+ * available data.
+ */
 void
 flatpak_transaction_set_no_pull (FlatpakTransaction *self,
                                  gboolean            no_pull)
@@ -845,6 +1016,14 @@ flatpak_transaction_set_no_pull (FlatpakTransaction *self,
   priv->no_pull = no_pull;
 }
 
+/**
+ * flatpak_transaction_set_no_deploy:
+ * @self: a #FlatpakTransaction
+ * @no_deploy: whether to avoid deploying
+ *
+ * Sets whether the transaction should download updates, but
+ * not deploy them.
+ */
 void
 flatpak_transaction_set_no_deploy (FlatpakTransaction *self,
                                    gboolean            no_deploy)
@@ -854,6 +1033,14 @@ flatpak_transaction_set_no_deploy (FlatpakTransaction *self,
   priv->no_deploy = no_deploy;
 }
 
+/**
+ * flatpak_transaction_set_disable_static_deltas:
+ * @self: a #FlatpakTransaction
+ * @disable_static_deltas: whether to avoid static deltas
+ *
+ * Sets whether the transaction should avoid using static
+ * deltas when pulling.
+ */
 void
 flatpak_transaction_set_disable_static_deltas (FlatpakTransaction *self,
                                                gboolean            disable_static_deltas)
@@ -863,6 +1050,14 @@ flatpak_transaction_set_disable_static_deltas (FlatpakTransaction *self,
   priv->disable_static_deltas = disable_static_deltas;
 }
 
+/**
+ * flatpak_transaction_set_disable_prune:
+ * @self: a #FlatpakTransaction
+ * @disable_prune: whether to avoid pruning
+ *
+ * Sets whether the transaction should avoid pruning the local OSTree
+ * repository after updating.
+ */
 void
 flatpak_transaction_set_disable_prune (FlatpakTransaction *self,
                                        gboolean            disable_prune)
@@ -872,6 +1067,14 @@ flatpak_transaction_set_disable_prune (FlatpakTransaction *self,
   priv->disable_prune = disable_prune;
 }
 
+/**
+ * flatpak_transaction_set_disable_dependencies:
+ * @self: a #FlatpakTransaction
+ * @disable_dependencies: whether to disable runtime dependencies
+ *
+ * Sets whether the transaction should ignore runtime dependencies
+ * when resolving operations for applications.
+ */
 void
 flatpak_transaction_set_disable_dependencies (FlatpakTransaction *self,
                                               gboolean            disable_dependencies)
@@ -881,6 +1084,15 @@ flatpak_transaction_set_disable_dependencies (FlatpakTransaction *self,
   priv->disable_deps = disable_dependencies;
 }
 
+/**
+ * flatpak_transaction_set_disable_related:
+ * @self: a #FlatpakTransaction
+ * @disable_related: whether to avoid adding related refs
+ *
+ * Sets whether the transaction should avoid adding related refs
+ * when resolving operations. Related refs are extensions that are
+ * suggested by apps, such as locales.
+ */
 void
 flatpak_transaction_set_disable_related (FlatpakTransaction *self,
                                          gboolean            disable_related)
@@ -890,6 +1102,14 @@ flatpak_transaction_set_disable_related (FlatpakTransaction *self,
   priv->disable_related = disable_related;
 }
 
+/**
+ * flatpak_transaction_set_reinstall:
+ * @self: a #FlatpakTransaction
+ * @reinstall: whether to reinstall refs
+ *
+ * Sets whether the transaction should uninstall first if a
+ * ref is already installed.
+ */
 void
 flatpak_transaction_set_reinstall (FlatpakTransaction *self,
                                    gboolean            reinstall)
@@ -899,6 +1119,14 @@ flatpak_transaction_set_reinstall (FlatpakTransaction *self,
   priv->reinstall = reinstall;
 }
 
+/**
+ * flatpak_transaction_set_force_uninstall:
+ * @self: a #FlatpakTransaction
+ * @force_uninstall: whether to force-uninstall refs
+ *
+ * Sets whether the transaction should uninstall files even
+ * if they're used by a running application.
+ */
 void
 flatpak_transaction_set_force_uninstall (FlatpakTransaction *self,
                                          gboolean            force_uninstall)
@@ -908,14 +1136,21 @@ flatpak_transaction_set_force_uninstall (FlatpakTransaction *self,
   priv->force_uninstall = force_uninstall;
 }
 
+/**
+ * flatpak_transaction_set_default_arch:
+ * @self: a #FlatpakTransaction
+ * @arch: the arch to make default
+ *
+ * Sets the architecture to default to where it is unspecified.
+ */ 
 void
 flatpak_transaction_set_default_arch (FlatpakTransaction *self,
-                                      const char         *default_arch)
+                                      const char         *arch)
 {
   FlatpakTransactionPrivate *priv = flatpak_transaction_get_instance_private (self);
 
   g_free (priv->default_arch);
-  priv->default_arch = g_strdup (default_arch);
+  priv->default_arch = g_strdup (arch);
 }
 
 static FlatpakTransactionOperation *
@@ -1178,13 +1413,13 @@ find_runtime_remote (FlatpakTransaction             *self,
   if (remotes == NULL || *remotes == NULL)
     {
       flatpak_fail_error (error, FLATPAK_ERROR_RUNTIME_NOT_FOUND,
-                          _("The Application %s requires the %s which was not found"),
+                          _("The application %s requires the runtime %s which was not found"),
                           app_pref, runtime_pref);
       return NULL;
     }
 
-  /* In the no-puil case, if only one local ref is available, assume that is the one becasue
-     the user chosed it interactively when pulling */
+  /* In the no-puil case, if only one local ref is available, assume that is the one because
+     the user chose it interactively when pulling */
   if (priv->no_pull && g_strv_length (remotes) == 1)
     res = 0;
   else
@@ -1194,7 +1429,7 @@ find_runtime_remote (FlatpakTransaction             *self,
     return g_strdup (remotes[res]);
 
   flatpak_fail_error (error, FLATPAK_ERROR_RUNTIME_NOT_FOUND,
-                      _("The Application %s requires the %s which is not installed"),
+                      _("The application %s requires the runtime %s which is not installed"),
                       app_pref, runtime_pref);
   return NULL;
 }
@@ -1327,11 +1562,8 @@ flatpak_transaction_add_ref (FlatpakTransaction             *self,
   if (kind == FLATPAK_TRANSACTION_OPERATION_UPDATE)
     {
       if (!dir_ref_is_installed (priv->dir, ref, &origin, NULL))
-        {
-          g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_NOT_INSTALLED,
-                       _("%s not installed"), pref);
-          return FALSE;
-        }
+        return flatpak_fail_error (error, FLATPAK_ERROR_NOT_INSTALLED,
+                                   _("%s not installed"), pref);
 
       if (flatpak_dir_get_remote_disabled (priv->dir, origin))
         {
@@ -1346,27 +1578,19 @@ flatpak_transaction_add_ref (FlatpakTransaction             *self,
           dir_ref_is_installed (priv->dir, ref, &origin, NULL))
         {
           if (g_strcmp0 (remote, origin) == 0)
-            {
-              g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
-                           _("%s is already installed"), pref);
-              return FALSE;
-            }
+            return flatpak_fail_error (error, FLATPAK_ERROR_ALREADY_INSTALLED,
+                                       _("%s is already installed"), pref);
           else
-            {
-              g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_DIFFERENT_REMOTE,
-                           _("%s is already installed from remote %s"), pref, origin);
-              return FALSE;
-            }
+            return flatpak_fail_error (error, FLATPAK_ERROR_DIFFERENT_REMOTE,
+                                       _("%s is already installed from remote %s"),
+                                       pref, origin);
         }
     }
   else if (kind == FLATPAK_TRANSACTION_OPERATION_UNINSTALL)
     {
       if (!dir_ref_is_installed (priv->dir, ref, &origin, NULL))
-        {
-          g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_NOT_INSTALLED,
-                       _("%s not installed"), pref);
-          return FALSE;
-        }
+        flatpak_fail_error (error, FLATPAK_ERROR_NOT_INSTALLED,
+                            _("%s not installed"), pref);
 
       remote = origin;
     }
@@ -1442,6 +1666,16 @@ flatpak_transaction_add_install_bundle (FlatpakTransaction *self,
   return TRUE;
 }
 
+/**
+ * flatpak_transaction_add_install_flatpakref:
+ * @self: a #FlatpakTransaction
+ * @flatpakref_data: data from a flatpakref file
+ * @error: return location for a #GError
+ *
+ * Adds installing the given flatpakref to this transaction.
+ *
+ * Returns: %TRUE on success; %FALSE with @error set on failure.
+ */
 gboolean
 flatpak_transaction_add_install_flatpakref (FlatpakTransaction *self,
                                             GBytes             *flatpakref_data,
@@ -1457,7 +1691,7 @@ flatpak_transaction_add_install_flatpakref (FlatpakTransaction *self,
   if (!g_key_file_load_from_data (keyfile, g_bytes_get_data (flatpakref_data, NULL),
                                   g_bytes_get_size (flatpakref_data),
                                   0, &local_error))
-    return flatpak_fail (error, "Invalid .flatpakref: %s", local_error->message);
+    return flatpak_fail_error (error, FLATPAK_ERROR_INVALID_DATA, _("Invalid .flatpakref: %s"), local_error->message);
 
   priv->flatpakrefs = g_list_append (priv->flatpakrefs, g_steal_pointer (&keyfile));
 
@@ -2098,7 +2332,7 @@ handle_runtime_repo_deps (FlatpakTransaction *self, const char *id, const char *
                                   g_bytes_get_data (dep_data, NULL),
                                   g_bytes_get_size (dep_data),
                                   0, error))
-    return flatpak_fail (error, "Invalid .flatpakrepo: %s", local_error->message);
+    return flatpak_fail_error (error, FLATPAK_ERROR_INVALID_DATA, _("Invalid .flatpakrepo: %s"), local_error->message);
 
   uri = soup_uri_new (dep_url);
   basename = g_path_get_basename (soup_uri_get_path (uri));
@@ -2291,6 +2525,27 @@ flatpak_transaction_resolve_bundles (FlatpakTransaction *self,
   return TRUE;
 }
 
+/**
+ * flatpak_transaction_run:
+ * @self: a #FlatpakTransaction
+ * @cancellable: (nullable): a #GCancellable
+ * @error: return location for an error
+ *
+ * Executes the transaction.
+ *
+ * During the cause of the execution, various signal will get emitted.
+ * The FlatpakTransaction::choose-remote-for-ref  and
+ * #FlatpakTransaction::add-new-remote signals may get emitted while
+ * resolving operations. #FlatpakTransaction::ready is emitted when
+ * the transaction has been fully resolved, and #FlatpakTransaction::new-operation
+ * and #FlatpakTransaction::operation-done are emitted while the operations
+ * are carried out. If an error occurs at any point during the execution,
+ * #FlatpakTransaction::operation-error is emitted.
+ *
+ * Note that this call blocks until the transaction is done.
+ *
+ * Returns: %TRUE on success, %FALSE if an error occurred
+ */
 gboolean
 flatpak_transaction_run (FlatpakTransaction *self,
                          GCancellable       *cancellable,
@@ -2386,10 +2641,7 @@ flatpak_transaction_run (FlatpakTransaction *self,
 
   g_signal_emit (self, signals[READY], 0, &ready_res);
   if (!ready_res)
-    {
-      g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_ABORTED, _("Aborted by user"));
-      return FALSE;
-    }
+    return flatpak_fail_error (error, FLATPAK_ERROR_ABORTED, _("Aborted by user"));
 
   for (l = priv->ops; l != NULL; l = l->next)
     {
@@ -2414,8 +2666,8 @@ flatpak_transaction_run (FlatpakTransaction *self,
            * remote is fixed. */
           !(op->fail_if_op_fails->kind == FLATPAK_TRANSACTION_OPERATION_UPDATE && g_str_has_prefix (op->ref, "app/")))
         {
-          g_set_error (&local_error, FLATPAK_ERROR, FLATPAK_ERROR_SKIPPED,
-                       _("Skipping %s due to previous error"), pref);
+          flatpak_fail_error (&local_error, FLATPAK_ERROR_SKIPPED,
+                              _("Skipping %s due to previous error"), pref);
           res = FALSE;
         }
       else if ((state = flatpak_transaction_ensure_remote_state (self, op->kind, op->remote, &local_error)) == NULL)
@@ -2588,8 +2840,13 @@ flatpak_transaction_run (FlatpakTransaction *self,
 
           if (!do_cont)
             {
-              g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_ABORTED,
-                           _("Aborted due to failure"));
+              if (g_cancellable_set_error_if_cancelled (cancellable, error))
+                {
+                  succeeded = FALSE;
+                  break;
+                }
+
+              flatpak_fail_error (error, FLATPAK_ERROR_ABORTED, _("Aborted due to failure"));
               succeeded = FALSE;
               break;
             }
