@@ -26,10 +26,10 @@
 #include <gio/gio.h>
 #include <polkit/polkit.h>
 
-#include "flatpak-dbus.h"
-#include "flatpak-dir.h"
-#include "flatpak-oci-registry.h"
-#include "lib/flatpak-error.h"
+#include "flatpak-dbus-generated.h"
+#include "flatpak-dir-private.h"
+#include "flatpak-oci-registry-private.h"
+#include "flatpak-error.h"
 
 static PolkitAuthority *authority = NULL;
 static FlatpakSystemHelper *helper = NULL;
@@ -39,13 +39,13 @@ static guint name_owner_id = 0;
 static gboolean on_session_bus = FALSE;
 static gboolean no_idle_exit = FALSE;
 
-#define IDLE_TIMEOUT_SECS 10*60
+#define IDLE_TIMEOUT_SECS 10 * 60
 
 /* This uses a weird Auto prefix to avoid conflicts with later added polkit types.
  */
 typedef PolkitAuthorizationResult AutoPolkitAuthorizationResult;
-typedef PolkitDetails AutoPolkitDetails;
-typedef PolkitSubject AutoPolkitSubject;
+typedef PolkitDetails             AutoPolkitDetails;
+typedef PolkitSubject             AutoPolkitSubject;
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (AutoPolkitAuthorizationResult, g_object_unref)
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (AutoPolkitDetails, g_object_unref)
@@ -62,6 +62,7 @@ static gboolean
 unref_skeleton_in_timeout_cb (gpointer user_data)
 {
   static gboolean unreffed = FALSE;
+
   g_debug ("unreffing helper main ref");
   if (!unreffed)
     {
@@ -104,7 +105,7 @@ schedule_idle_callback (void)
 {
   static guint idle_timeout_id = 0;
 
-  G_LOCK(idle);
+  G_LOCK (idle);
 
   if (!no_idle_exit)
     {
@@ -114,7 +115,7 @@ schedule_idle_callback (void)
       idle_timeout_id = g_timeout_add_seconds (IDLE_TIMEOUT_SECS, idle_timeout_cb, NULL);
     }
 
-  G_UNLOCK(idle);
+  G_UNLOCK (idle);
 }
 
 static FlatpakDir *
@@ -160,7 +161,6 @@ handle_deploy (FlatpakSystemHelper   *object,
   gboolean no_deploy;
   gboolean local_pull;
   gboolean reinstall;
-  g_autoptr(GMainContext) main_context = NULL;
   g_autofree char *url = NULL;
 
   g_debug ("Deploy %s %u %s %s %s", arg_repo_path, arg_flags, arg_ref, arg_origin, arg_installation);
@@ -219,7 +219,7 @@ handle_deploy (FlatpakSystemHelper   *object,
   else if (!deploy_dir && is_update)
     {
       /* Can't update not installed app */
-      g_dbus_method_invocation_return_error (invocation, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
+      g_dbus_method_invocation_return_error (invocation, FLATPAK_ERROR, FLATPAK_ERROR_NOT_INSTALLED,
                                              "%s is not installed", arg_ref);
       return TRUE;
     }
@@ -241,9 +241,12 @@ handle_deploy (FlatpakSystemHelper   *object,
       g_autoptr(FlatpakOciIndex) index = NULL;
       const FlatpakOciManifestDescriptor *desc;
       g_autoptr(FlatpakOciVersioned) versioned = NULL;
+      g_autoptr(FlatpakRemoteState) state = NULL;
+      FlatpakCollectionRef collection_ref;
+      g_autoptr(GHashTable) remote_refs = NULL;
       g_autofree char *checksum = NULL;
+      const char *verified_digest;
       g_autofree char *upstream_url = NULL;
-      g_autoptr(SoupSession) soup_session = NULL;
 
       ostree_repo_remote_get_url (flatpak_dir_get_repo (system),
                                   arg_origin,
@@ -251,11 +254,11 @@ handle_deploy (FlatpakSystemHelper   *object,
                                   NULL);
 
       if (upstream_url == NULL)
-	{
-	  g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-						 "Remote %s is disabled", arg_origin);
-	  return TRUE;
-	}
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Remote %s is disabled", arg_origin);
+          return TRUE;
+        }
 
       registry = flatpak_oci_registry_new (registry_uri, FALSE, -1, NULL, &error);
       if (registry == NULL)
@@ -265,7 +268,7 @@ handle_deploy (FlatpakSystemHelper   *object,
           return TRUE;
         }
 
-      index = flatpak_oci_registry_load_index (registry, NULL, NULL, NULL, &error);
+      index = flatpak_oci_registry_load_index (registry, NULL, &error);
       if (index == NULL)
         {
           g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
@@ -290,16 +293,42 @@ handle_deploy (FlatpakSystemHelper   *object,
           return TRUE;
         }
 
-      soup_session = flatpak_create_soup_session (PACKAGE_STRING);
-      if (!flatpak_oci_index_verify_ref (soup_session,
-					 upstream_url,
-					 arg_ref,
-					 desc->parent.digest,
-					 NULL, &error))
-	{
-	  g_dbus_method_invocation_return_gerror (invocation, error);
-	  return TRUE;
-	}
+      state = flatpak_dir_get_remote_state (system, arg_origin, NULL, &error);
+      if (state == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "%s: Can't get remote state: %s", arg_origin, error->message);
+          return TRUE;
+        }
+
+      /* We need to use list_all_remote_refs because we don't care about
+       * enumerate vs. noenumerate.
+       */
+      if (!flatpak_dir_list_all_remote_refs (system, state, &remote_refs, NULL, &error))
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "%s: Can't list refs: %s", arg_origin, error->message);
+          return TRUE;
+        }
+
+      collection_ref.collection_id = state->collection_id;
+      collection_ref.ref_name = (char *) arg_ref;
+
+      verified_digest = g_hash_table_lookup (remote_refs, &collection_ref);
+      if (!verified_digest)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "%s: ref %s not found", arg_origin, arg_ref);
+          return TRUE;
+        }
+
+      if (!g_str_has_prefix (desc->parent.digest, "sha256:") ||
+          strcmp (desc->parent.digest + strlen ("sha256:"), verified_digest) != 0)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "%s: manifest hash in downloaded content does not match ref %s", arg_origin, arg_ref);
+          return TRUE;
+        }
 
       checksum = flatpak_pull_from_oci (flatpak_dir_get_repo (system), registry, NULL, desc->parent.digest, FLATPAK_OCI_MANIFEST (versioned),
                                         arg_origin, arg_ref, NULL, NULL, NULL, &error);
@@ -312,9 +341,10 @@ handle_deploy (FlatpakSystemHelper   *object,
     }
   else if (strlen (arg_repo_path) > 0)
     {
+      g_autoptr(GMainContextPopDefault) main_context = NULL;
+
       /* Work around ostree-pull spinning the default main context for the sync calls */
-      main_context = g_main_context_new ();
-      g_main_context_push_thread_default (main_context);
+      main_context = flatpak_main_context_new_default ();
 
       ostree_progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
 
@@ -325,18 +355,23 @@ handle_deploy (FlatpakSystemHelper   *object,
                                              ostree_progress,
                                              NULL, &error))
         {
-          g_main_context_pop_thread_default (main_context);
-          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                                 "Error pulling from repo: %s", error->message);
+          if (error->domain == FLATPAK_ERROR)
+            g_dbus_method_invocation_return_gerror (invocation, error);
+          else
+            g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                   "Error pulling from repo: %s", error->message);
+
           return TRUE;
         }
-      g_main_context_pop_thread_default (main_context);
 
       if (ostree_progress)
         ostree_async_progress_finish (ostree_progress);
     }
   else if (local_pull)
     {
+      g_autoptr(GMainContextPopDefault) main_context = NULL;
+
+      g_autoptr(FlatpakRemoteState) state = NULL;
       if (!ostree_repo_remote_get_url (flatpak_dir_get_repo (system),
                                        arg_origin,
                                        &url,
@@ -354,23 +389,31 @@ handle_deploy (FlatpakSystemHelper   *object,
           return TRUE;
         }
 
-      /* Work around ostree-pull spinning the default main context for the sync calls */
-      main_context = g_main_context_new ();
-      g_main_context_push_thread_default (main_context);
-
-      ostree_progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
-
-      if (!flatpak_dir_pull (system, arg_origin, arg_ref, NULL, NULL, (const char **)arg_subpaths, NULL,
-                             FLATPAK_PULL_FLAGS_NONE, OSTREE_REPO_PULL_FLAGS_UNTRUSTED, ostree_progress,
-                             NULL, &error))
+      state = flatpak_dir_get_remote_state_optional (system, arg_origin, NULL, &error);
+      if (state == NULL)
         {
-          g_main_context_pop_thread_default (main_context);
           g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
                                                  "Error pulling from repo: %s", error->message);
           return TRUE;
         }
 
-      g_main_context_pop_thread_default (main_context);
+      /* Work around ostree-pull spinning the default main context for the sync calls */
+      main_context = flatpak_main_context_new_default ();
+
+      ostree_progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
+
+      if (!flatpak_dir_pull (system, state, arg_ref, NULL, NULL, (const char **) arg_subpaths, NULL,
+                             FLATPAK_PULL_FLAGS_NONE, OSTREE_REPO_PULL_FLAGS_UNTRUSTED, ostree_progress,
+                             NULL, &error))
+        {
+          if (error->domain == FLATPAK_ERROR)
+            g_dbus_method_invocation_return_gerror (invocation, error);
+          else
+            g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                   "Error pulling from repo: %s", error->message);
+
+          return TRUE;
+        }
 
       if (ostree_progress)
         ostree_async_progress_finish (ostree_progress);
@@ -381,7 +424,7 @@ handle_deploy (FlatpakSystemHelper   *object,
       if (is_update)
         {
           if (!flatpak_dir_deploy_update (system, arg_ref,
-                                          NULL, (const char **)arg_subpaths, NULL, &error))
+                                          NULL, (const char **) arg_subpaths, NULL, &error))
             {
               if (g_error_matches (error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED))
                 g_dbus_method_invocation_return_error (invocation, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
@@ -420,10 +463,9 @@ handle_deploy_appstream (FlatpakSystemHelper   *object,
                          const gchar           *arg_installation)
 {
   g_autoptr(FlatpakDir) system = NULL;
-  g_autoptr(GFile) path = g_file_new_for_path (arg_repo_path);
   g_autoptr(GError) error = NULL;
-  g_autoptr(GMainContext) main_context = NULL;
-  g_autofree char *branch = NULL;
+  g_autofree char *new_branch = NULL;
+  g_autofree char *old_branch = NULL;
   gboolean is_oci;
 
   g_debug ("DeployAppstream %s %s %s %s", arg_repo_path, arg_origin, arg_arch, arg_installation);
@@ -435,10 +477,14 @@ handle_deploy_appstream (FlatpakSystemHelper   *object,
       return TRUE;
     }
 
-  if (!g_file_query_exists (path, NULL))
+  if (strlen (arg_repo_path) > 0)
     {
-      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "Path does not exist");
-      return TRUE;
+      g_autoptr(GFile) path = g_file_new_for_path (arg_repo_path);
+      if (!g_file_query_exists (path, NULL))
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "Path does not exist");
+          return TRUE;
+        }
     }
 
   if (!flatpak_dir_ensure_repo (system, NULL, &error))
@@ -450,80 +496,123 @@ handle_deploy_appstream (FlatpakSystemHelper   *object,
 
   is_oci = flatpak_dir_get_remote_oci (system, arg_origin);
 
-  branch = g_strdup_printf ("appstream/%s", arg_arch);
+  new_branch = g_strdup_printf ("appstream2/%s", arg_arch);
+  old_branch = g_strdup_printf ("appstream/%s", arg_arch);
 
   if (is_oci)
     {
-      g_autoptr(GFile) registry_file = g_file_new_for_path (arg_repo_path);
-      g_autofree char *registry_uri = g_file_get_uri (registry_file);
-      g_autoptr(FlatpakOciRegistry) registry = NULL;
-      g_autoptr(FlatpakOciIndex) index = NULL;
-      const FlatpakOciManifestDescriptor *desc;
-      g_autoptr(FlatpakOciVersioned) versioned = NULL;
-      g_autofree char *checksum = NULL;
+      g_autoptr(GMainContextPopDefault) context =  NULL;
 
-      registry = flatpak_oci_registry_new (registry_uri, FALSE, -1, NULL, &error);
-      if (registry == NULL)
+      /* This does soup http requests spinning the current mainloop, so we need one
+         for this thread. */
+      context = flatpak_main_context_new_default ();
+      /* In the OCI case, we just do the full update, including network i/o, in the
+       * system helper, see comment in flatpak_dir_update_appstream()
+       */
+      if (!flatpak_dir_update_appstream (system,
+                                         arg_origin,
+                                         arg_arch,
+                                         NULL,
+                                         NULL,
+                                         NULL,
+                                         &error))
         {
           g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                                 "Can't open child OCI registry: %s", error->message);
+                                                 "Error updating appstream: %s", error->message);
           return TRUE;
         }
 
-      index = flatpak_oci_registry_load_index (registry, NULL, NULL, NULL, &error);
-      if (index == NULL)
-        {
-          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                                 "Can't open child OCI registry index: %s", error->message);
-          return TRUE;
-        }
-
-      desc = flatpak_oci_index_get_manifest (index, branch);
-      if (desc == NULL)
-        {
-          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                                 "Can't find ref %s in child OCI registry index", branch);
-          return TRUE;
-        }
-
-      versioned = flatpak_oci_registry_load_versioned (registry, NULL, desc->parent.digest, NULL,
-                                                       NULL, &error);
-      if (versioned == NULL || !FLATPAK_IS_OCI_MANIFEST (versioned))
-        {
-          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                                 "Can't open child manifest");
-          return TRUE;
-        }
-
-      checksum = flatpak_pull_from_oci (flatpak_dir_get_repo (system), registry, NULL, desc->parent.digest, FLATPAK_OCI_MANIFEST (versioned),
-                                        arg_origin, branch, NULL, NULL, NULL, &error);
-      if (checksum == NULL)
-        {
-          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                                 "Can't pull ref %s from child OCI registry index: %s", branch, error->message);
-          return TRUE;
-        }
+      flatpak_system_helper_complete_deploy_appstream (object, invocation);
+      return TRUE;
     }
-  else
+  else if (strlen (arg_repo_path) > 0)
     {
+      g_autoptr(GError) first_error = NULL;
+      g_autoptr(GError) second_error = NULL;
+      g_autoptr(GMainContextPopDefault) main_context = NULL;
+
       /* Work around ostree-pull spinning the default main context for the sync calls */
-      main_context = g_main_context_new ();
-      g_main_context_push_thread_default (main_context);
+      main_context = flatpak_main_context_new_default ();
 
       if (!flatpak_dir_pull_untrusted_local (system, arg_repo_path,
                                              arg_origin,
-                                             branch,
+                                             new_branch,
                                              NULL,
                                              NULL,
-                                             NULL, &error))
+                                             NULL, &first_error))
+        {
+          if (!flatpak_dir_pull_untrusted_local (system, arg_repo_path,
+                                                 arg_origin,
+                                                 old_branch,
+                                                 NULL,
+                                                 NULL,
+                                                 NULL, &second_error))
+            {
+              g_prefix_error (&first_error, "Error updating appstream2: ");
+              g_prefix_error (&second_error, "%s; Error updating appstream: ", first_error->message);
+              g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                     "Error pulling from repo: %s", second_error->message);
+              return TRUE;
+            }
+        }
+    }
+  else /* empty path == local pull */
+    {
+      g_autoptr(FlatpakRemoteState) state = NULL;
+      g_autoptr(OstreeAsyncProgress) ostree_progress = NULL;
+      g_autoptr(GError) first_error = NULL;
+      g_autoptr(GError) second_error = NULL;
+      g_autofree char *url = NULL;
+      g_autoptr(GMainContextPopDefault) main_context = NULL;
+
+      if (!ostree_repo_remote_get_url (flatpak_dir_get_repo (system),
+                                       arg_origin,
+                                       &url,
+                                       &error))
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Error getting remote url: %s", error->message);
+          return TRUE;
+        }
+
+      if (!g_str_has_prefix (url, "file:"))
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Local pull url doesn't start with file://");
+          return TRUE;
+        }
+
+      state = flatpak_dir_get_remote_state_optional (system, arg_origin, NULL, &error);
+      if (state == NULL)
         {
-          g_main_context_pop_thread_default (main_context);
           g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
                                                  "Error pulling from repo: %s", error->message);
           return TRUE;
         }
 
-      g_main_context_pop_thread_default (main_context);
+      /* Work around ostree-pull spinning the default main context for the sync calls */
+      main_context = flatpak_main_context_new_default ();
+
+      ostree_progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
+
+      if (!flatpak_dir_pull (system, state, new_branch, NULL, NULL, NULL, NULL,
+                             FLATPAK_PULL_FLAGS_NONE, OSTREE_REPO_PULL_FLAGS_UNTRUSTED, ostree_progress,
+                             NULL, &first_error))
+        {
+          if (!flatpak_dir_pull (system, state, old_branch, NULL, NULL, NULL, NULL,
+                                 FLATPAK_PULL_FLAGS_NONE, OSTREE_REPO_PULL_FLAGS_UNTRUSTED, ostree_progress,
+                                 NULL, &second_error))
+            {
+              g_prefix_error (&first_error, "Error updating appstream2: ");
+              g_prefix_error (&second_error, "%s; Error updating appstream: ", first_error->message);
+              g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                     "Error pulling from repo: %s", second_error->message);
+              return TRUE;
+            }
+        }
+
+      if (ostree_progress)
+        ostree_async_progress_finish (ostree_progress);
     }
 
   if (!flatpak_dir_deploy_appstream (system,
@@ -544,11 +633,11 @@ handle_deploy_appstream (FlatpakSystemHelper   *object,
 }
 
 static gboolean
-handle_uninstall (FlatpakSystemHelper *object,
+handle_uninstall (FlatpakSystemHelper   *object,
                   GDBusMethodInvocation *invocation,
-                  guint arg_flags,
-                  const gchar *arg_ref,
-                  const gchar *arg_installation)
+                  guint                  arg_flags,
+                  const gchar           *arg_ref,
+                  const gchar           *arg_installation)
 {
   g_autoptr(FlatpakDir) system = NULL;
   g_autoptr(GError) error = NULL;
@@ -635,13 +724,13 @@ handle_install_bundle (FlatpakSystemHelper   *object,
 
 
 static gboolean
-handle_configure_remote (FlatpakSystemHelper *object,
+handle_configure_remote (FlatpakSystemHelper   *object,
                          GDBusMethodInvocation *invocation,
-                         guint arg_flags,
-                         const gchar *arg_remote,
-                         const gchar *arg_config,
-                         GVariant *arg_gpg_key,
-                         const gchar *arg_installation)
+                         guint                  arg_flags,
+                         const gchar           *arg_remote,
+                         const gchar           *arg_config,
+                         GVariant              *arg_gpg_key,
+                         const gchar           *arg_installation)
 {
   g_autoptr(FlatpakDir) system = NULL;
   g_autoptr(GError) error = NULL;
@@ -721,12 +810,12 @@ handle_configure_remote (FlatpakSystemHelper *object,
 }
 
 static gboolean
-handle_configure (FlatpakSystemHelper *object,
-		  GDBusMethodInvocation *invocation,
-		  guint arg_flags,
-		  const gchar *arg_key,
-		  const gchar *arg_value,
-		  const gchar *arg_installation)
+handle_configure (FlatpakSystemHelper   *object,
+                  GDBusMethodInvocation *invocation,
+                  guint                  arg_flags,
+                  const gchar           *arg_key,
+                  const gchar           *arg_value,
+                  const gchar           *arg_installation)
 {
   g_autoptr(FlatpakDir) system = NULL;
   g_autoptr(GError) error = NULL;
@@ -776,25 +865,23 @@ handle_configure (FlatpakSystemHelper *object,
 }
 
 static gboolean
-handle_update_remote (FlatpakSystemHelper *object,
+handle_update_remote (FlatpakSystemHelper   *object,
                       GDBusMethodInvocation *invocation,
-                      guint arg_flags,
-                      const gchar *arg_remote,
-                      const gchar *arg_installation,
-                      const gchar *arg_summary_path,
-                      const gchar *arg_summary_sig_path)
+                      guint                  arg_flags,
+                      const gchar           *arg_remote,
+                      const gchar           *arg_installation,
+                      const gchar           *arg_summary_path,
+                      const gchar           *arg_summary_sig_path)
 {
   g_autoptr(FlatpakDir) system = NULL;
   g_autoptr(GError) error = NULL;
   char *summary_data = NULL;
   gsize summary_size;
   g_autoptr(GBytes) summary_bytes = NULL;
-  g_autoptr(GVariant) summary = NULL;
   char *summary_sig_data = NULL;
   gsize summary_sig_size;
   g_autoptr(GBytes) summary_sig_bytes = NULL;
-  g_autoptr(OstreeGpgVerifyResult) gpg_result = NULL;
-  g_autofree char *collection_id = NULL;
+  g_autoptr(FlatpakRemoteState) state = NULL;
 
   g_debug ("UpdateRemote %u %s %s %s %s", arg_flags, arg_remote, arg_installation, arg_summary_path, arg_summary_sig_path);
 
@@ -819,23 +906,6 @@ handle_update_remote (FlatpakSystemHelper *object,
       return TRUE;
     }
 
-  if (!flatpak_dir_ensure_repo (system, NULL, &error))
-    {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
-    }
-
-#ifdef FLATPAK_ENABLE_P2P
-  if (!ostree_repo_get_remote_option (flatpak_dir_get_repo (system), arg_remote, "collection-id",
-                                      NULL, &collection_id, &error))
-    {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
-    }
-  else if (collection_id != NULL && *collection_id == '\0')
-    g_clear_pointer (&collection_id, g_free);
-#endif  /* FLATPAK_ENABLE_P2P */
-
   if (!g_file_get_contents (arg_summary_path, &summary_data, &summary_size, &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -843,7 +913,7 @@ handle_update_remote (FlatpakSystemHelper *object,
     }
   summary_bytes = g_bytes_new_take (summary_data, summary_size);
 
-  if (collection_id == NULL)
+  if (*arg_summary_sig_path != 0)
     {
       if (!g_file_get_contents (arg_summary_sig_path, &summary_sig_data, &summary_sig_size, &error))
         {
@@ -851,36 +921,26 @@ handle_update_remote (FlatpakSystemHelper *object,
           return TRUE;
         }
       summary_sig_bytes = g_bytes_new_take (summary_sig_data, summary_sig_size);
-
-      gpg_result = ostree_repo_verify_summary (flatpak_dir_get_repo (system),
-                                               arg_remote,
-                                               summary_bytes,
-                                               summary_sig_bytes,
-                                               NULL, &error);
-      if (gpg_result == NULL ||
-          !ostree_gpg_verify_result_require_valid_signature (gpg_result, &error))
-        {
-          g_dbus_method_invocation_return_gerror (invocation, error);
-          return TRUE;
-        }
-
-      summary = g_variant_ref_sink (g_variant_new_from_bytes (OSTREE_SUMMARY_GVARIANT_FORMAT,
-                                                              summary_bytes, FALSE));
-      if (!flatpak_dir_update_remote_configuration_for_summary (system, arg_remote, summary,
-                                                                FALSE, NULL, NULL, &error))
-        {
-          g_dbus_method_invocation_return_gerror (invocation, error);
-          return TRUE;
-        }
     }
-  else
+
+  state = flatpak_dir_get_remote_state_for_summary (system, arg_remote, summary_bytes, summary_sig_bytes, NULL, &error);
+  if (state == NULL)
     {
-      if (!flatpak_dir_update_remote_configuration_for_repo_metadata (system, arg_remote, summary,
-                                                                      FALSE, NULL, NULL, &error))
-        {
-          g_dbus_method_invocation_return_gerror (invocation, error);
-          return TRUE;
-        }
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (summary_sig_bytes == NULL && state->collection_id == NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                             "UpdateRemote requires a summary signature");
+      return TRUE;
+    }
+
+  if (!flatpak_dir_update_remote_configuration_for_state (system, state, FALSE, NULL, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
     }
 
   flatpak_system_helper_complete_update_remote (object, invocation);
@@ -965,6 +1025,103 @@ handle_prune_local_repo (FlatpakSystemHelper   *object,
   return TRUE;
 }
 
+
+static gboolean
+handle_ensure_repo (FlatpakSystemHelper   *object,
+                    GDBusMethodInvocation *invocation,
+                    const gchar           *arg_installation)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+
+  g_debug ("EnsureRepo %s", arg_installation);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  flatpak_system_helper_complete_ensure_repo (object, invocation);
+
+  return TRUE;
+}
+
+static gboolean
+handle_run_triggers (FlatpakSystemHelper   *object,
+                     GDBusMethodInvocation *invocation,
+                     const gchar           *arg_installation)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+
+  g_debug ("RunTriggers %s", arg_installation);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_run_triggers (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  flatpak_system_helper_complete_run_triggers (object, invocation);
+
+  return TRUE;
+}
+
+static gboolean
+handle_update_summary (FlatpakSystemHelper   *object,
+                       GDBusMethodInvocation *invocation,
+                       const gchar           *arg_installation)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+
+  g_debug ("UpdateSummary %s", arg_installation);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_update_summary (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  flatpak_system_helper_complete_update_summary (object, invocation);
+
+  return TRUE;
+}
+
 static gboolean
 flatpak_authorize_method_handler (GDBusInterfaceSkeleton *interface,
                                   GDBusMethodInvocation  *invocation,
@@ -973,6 +1130,7 @@ flatpak_authorize_method_handler (GDBusInterfaceSkeleton *interface,
   const gchar *method_name = g_dbus_method_invocation_get_method_name (invocation);
   const gchar *sender = g_dbus_method_invocation_get_sender (invocation);
   GVariant *parameters = g_dbus_method_invocation_get_parameters (invocation);
+
   g_autoptr(AutoPolkitSubject) subject = polkit_system_bus_name_new (sender);
   g_autoptr(AutoPolkitDetails) details = polkit_details_new ();
   const gchar *action = NULL;
@@ -1089,7 +1247,10 @@ flatpak_authorize_method_handler (GDBusInterfaceSkeleton *interface,
       polkit_details_insert (details, "remote", remote);
     }
   else if (g_strcmp0 (method_name, "RemoveLocalRef") == 0 ||
-           g_strcmp0 (method_name, "PruneLocalRepo") == 0)
+           g_strcmp0 (method_name, "PruneLocalRepo") == 0 ||
+           g_strcmp0 (method_name, "EnsureRepo") == 0 ||
+           g_strcmp0 (method_name, "RunTriggers") == 0 ||
+           g_strcmp0 (method_name, "UpdateSummary") == 0)
     {
       const char *remote;
 
@@ -1143,7 +1304,7 @@ on_bus_acquired (GDBusConnection *connection,
 
   flatpak_system_helper_set_version (FLATPAK_SYSTEM_HELPER (helper), 1);
 
-  g_object_set_data_full (G_OBJECT(helper), "track-alive", GINT_TO_POINTER(42), skeleton_died_cb);
+  g_object_set_data_full (G_OBJECT (helper), "track-alive", GINT_TO_POINTER (42), skeleton_died_cb);
 
   g_dbus_interface_skeleton_set_flags (G_DBUS_INTERFACE_SKELETON (helper),
                                        G_DBUS_INTERFACE_SKELETON_FLAGS_HANDLE_METHOD_INVOCATIONS_IN_THREAD);
@@ -1157,6 +1318,9 @@ on_bus_acquired (GDBusConnection *connection,
   g_signal_connect (helper, "handle-update-remote", G_CALLBACK (handle_update_remote), NULL);
   g_signal_connect (helper, "handle-remove-local-ref", G_CALLBACK (handle_remove_local_ref), NULL);
   g_signal_connect (helper, "handle-prune-local-repo", G_CALLBACK (handle_prune_local_repo), NULL);
+  g_signal_connect (helper, "handle-ensure-repo", G_CALLBACK (handle_ensure_repo), NULL);
+  g_signal_connect (helper, "handle-run-triggers", G_CALLBACK (handle_run_triggers), NULL);
+  g_signal_connect (helper, "handle-update-summary", G_CALLBACK (handle_update_summary), NULL);
 
   g_signal_connect (helper, "g-authorize-method",
                     G_CALLBACK (flatpak_authorize_method_handler),
@@ -1190,11 +1354,11 @@ on_name_lost (GDBusConnection *connection,
 }
 
 static void
-binary_file_changed_cb (GFileMonitor *file_monitor,
-                        GFile *file,
-                        GFile *other_file,
+binary_file_changed_cb (GFileMonitor     *file_monitor,
+                        GFile            *file,
+                        GFile            *other_file,
                         GFileMonitorEvent event_type,
-                        gpointer data)
+                        gpointer          data)
 {
   static gboolean got_it = FALSE;
 
@@ -1224,13 +1388,14 @@ int
 main (int    argc,
       char **argv)
 {
-  gchar exe_path[PATH_MAX+1];
+  gchar exe_path[PATH_MAX + 1];
   ssize_t exe_path_len;
   gboolean replace;
   gboolean verbose;
   gboolean show_version;
   GBusNameOwnerFlags flags;
   GOptionContext *context;
+
   g_autoptr(GError) error = NULL;
   const GOptionEntry options[] = {
     { "replace", 'r', 0, G_OPTION_ARG_NONE, &replace,  "Replace old daemon.", NULL },
@@ -1260,7 +1425,7 @@ main (int    argc,
 
   if (!g_option_context_parse (context, &argc, &argv, &error))
     {
-      g_printerr ("%s: %s", g_get_application_name(), error->message);
+      g_printerr ("%s: %s", g_get_application_name (), error->message);
       g_printerr ("\n");
       g_printerr ("Try \"%s --help\" for more information.",
                   g_get_prgname ());
